import { SdaDialogDelegate } from '../ppm/SdaDialogDelegate';
import { Log } from '../util/Log';
import { DialogProxyTools } from './DialogProxyTools';
import { ValueIterator } from './ValueIterator';
export class DialogProxy {
    constructor() {
        this._lastActivity = new Date();
        this._dialogDelegateChain = [new SdaDialogDelegate()];
    }
    get lastActivity() {
        return this._lastActivity;
    }
    getBlob(baseUrl, resourcePath) {
        return this.processRequestAndResponse('getBlob', 'handleGetBlobResponse', [baseUrl, resourcePath]);
    }
    getText(baseUrl, resourcePath) {
        return this.processRequestAndResponse('getText', 'handleGetTextResponse', [baseUrl, resourcePath]);
    }
    openStream(baseUrl, resourcePath) {
        return this.processRequestAndResponse('openStream', 'handleOpenStreamResponse', [baseUrl, resourcePath]);
    }
    postMultipart(baseUrl, resourcePath, formData) {
        return this.processRequestAndResponse('postMultipart', 'handlePostMultipartResponse', [
            baseUrl,
            resourcePath,
            formData
        ]);
    }
    getJson(baseUrl, resourcePath, queryParams) {
        return this.processRequestAndResponse('getJson', 'handleGetJsonResponse', [baseUrl, resourcePath, queryParams]);
    }
    postJson(baseUrl, resourcePath, jsonBody) {
        return this.processRequestAndResponse('postJson', 'handlePostJsonResponse', [baseUrl, resourcePath, jsonBody]);
    }
    putJson(baseUrl, resourcePath, jsonBody) {
        return this.processRequestAndResponse('putJson', 'handlePutJsonResponse', [baseUrl, resourcePath, jsonBody]);
    }
    deleteJson(baseUrl, resourcePath) {
        return this.processRequestAndResponse('deleteJson', 'handleDeleteJsonResponse', [baseUrl, resourcePath]);
    }
    static delegateRequest(previousPr, delegateIterator, requestFn, args) {
        const thisMethod = 'DialogProxy::delegateRequest';
        return previousPr.then(unusedValue => {
            if (delegateIterator.done()) {
                Log.info(`${thisMethod} -- using common fetch client to process request: ${requestFn}`);
                const fc = DialogProxyTools.commonFetchClient();
                return fc[requestFn].apply(fc, args);
            }
            // Select next delegate
            const nextDelegate = delegateIterator.next();
            const nextPr = nextDelegate[requestFn].apply(nextDelegate, args);
            if (!nextPr) {
                // Next delegate chose to immediately skip this request, so advance to the next delegate
                return this.delegateRequest(previousPr, delegateIterator, requestFn, args);
            }
            return nextPr.then(response => {
                if (!response) {
                    // Next delegate chose to skip this request after a delay, so advance to the next delegate
                    Log.info(`${thisMethod} -- delegate returned a falsey response, advancing to the next delegate with request: ${requestFn}`);
                    return this.delegateRequest(nextPr, delegateIterator, requestFn, args);
                }
                // Next delegate produced a response, so this is the future that will be processed
                return nextPr;
            });
        });
    }
    prepareForActivity() {
        this._lastActivity = new Date();
        if (!this._initialized) {
            Log.info('DialogProxy::prepareForActivity -- waiting for all DialogDelegates to initialize');
            const allDelegatesInitializing = this._dialogDelegateChain.map(d => d.initialize());
            this._initialized = Promise.all(allDelegatesInitializing).then(() => {
                Log.info('DialogProxy::prepareForActivity -- all DialogDelegates are initialized');
                Log.info('DialogProxy::prepareForActivity -- DialogProxy is initialized');
                return true;
            });
        }
    }
    processRequestAndResponse(requestFn, responseFn, args) {
        this.prepareForActivity();
        return this._initialized.then(() => {
            const delegateIterator = new ValueIterator(this._dialogDelegateChain);
            let responsePr = DialogProxy.delegateRequest(Promise.resolve(), delegateIterator, requestFn, args);
            for (const d of this._dialogDelegateChain) {
                const argsWithResponse = args.slice(0);
                argsWithResponse.push(responsePr);
                responsePr = d[responseFn].apply(d, argsWithResponse);
            }
            return responsePr;
        });
    }
}
